#!/usr/bin/python
# -*- coding:utf-8 -*-

#  Copyright © 2011, 2013  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function, division

import os, sys
import warnings
import Queue
from threading import Thread, current_thread
import cPickle as pickle

from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gtk

from launchpadlib.launchpad import Launchpad
from launchpadlib.errors import HTTPError
import lazr.restfulclient.resource


# See: https://help.launchpad.net/API/launchpadlib
#attrtypes = {
#    'lp_attributes':    ('attributes', 'Data fields of this object.'
#                            ' You can read from these might be able to write to some of them.'),
#    'lp_collections':   ('collections', 'List of launchpad objects associated with this object.'),
#    'lp_entries':       ('entries', 'Other Launchpad objects associated with this one.'),
#    'lp_operations':    ('operations', 'The names of Launchpad methods you can call on the object.'),
#}

# Definition of cached rows
(ROW_OK, ROW_LOADING, ROW_LINK, ROW_ERROR, ROW_AUTHENTICATION,) = range(5)  # Row.state
class Row (object):
    def __init__(self, parent=None, value=None, attrname='(empty)', attrtype=' ',
                        n_children=0, next=None, first_child=None, self_link='',
                        long_value=False, display_value='', state=ROW_OK, index=0):
        self.parent = parent                # parent Row object
        self.index = index
        self.n_children = n_children        # count children
        self.next = next                    # next row or None
        self.first_child = first_child      # first child row or None
        self.attrname = attrname            # attribute name
        self.attrtype = attrtype            # attribute type
        self.value = value                  # lp_object or value or None
        self.self_link = self_link          # == value.self_link
        self.long_value = long_value        # value has text, that should displayed multiline
        self.display_value = display_value  # value as text
        self.state = state
        self.error = ''                     # error string

    def delete(self):
        self.__init__()

    def __str__(self):
        return 'Row({0})'.format(',\n    '.join(['{0}={1!r}'.format(*i) for i in self.__dict__.items()]))

    def __getstate__(self):
        odict = self.__dict__.copy()
        odict['value'] = None
        if self.parent is not None:
            odict['parent'] = id(self.parent)
        if self.next is not None:
            odict['next'] = id(self.next)
        if self.first_child is not None:
            odict['first_child'] = id(self.first_child)
        if self.attrtype == 'O':
            odict['state'] = ROW_LOADING
        return odict

    # Definition of the model
    long_value = True
    show_hidden = False
    invisible_attrs = [
        'resource_type_link', 'http_etag',
        'entries', 'entry_links', 'self_link', 'prev', 'web',
    ]
    default_attrs = ['display_name', 'displayname']
    row_icons = ['gtk-yes', 'gtk-refresh', 'gtk-jump-to', 'gtk-dialog-error', 'gtk-dialog-authentication']
    # get column values from a cached row
    def is_visible(self):
        return (self.attrname not in Row.invisible_attrs or Row.show_hidden) and self.attrname != '(empty)'

    priority_attrtype = dict((c, i) for i, c in enumerate('AEOC '))
    def priority(self):
        pri = self.priority_attrtype[self.attrtype] + 10
        if self.attrname in ['start', 'total_size']:
            pri += 10
        return pri

    def get_row(self, path):
        for i in path:
            self = self.first_child
            for j in xrange(i):
                self = self.next
        return self

    def get_attr_path(self):
        attrnames = []
        while self is not None:
            attrnames.append(self.attrname)
            self = self.parent
        return '.'.join(reversed(attrnames))

dummy_row = Row()

class LPWorker (object):
    def __init__(self, launchpad, queue_size_notify_func):
        self.launchpad = launchpad
        self.queue_size_notify_func = queue_size_notify_func
        
        self.queue = Queue.PriorityQueue()
        self.worker = Thread(target=self._thread_loop)
        self.worker.daemon = True
        self.worker.start()
        
    def _thread_loop(self):
        while True:
            try:
                func, args, result_func = self.queue.get()[1]
                if func is None:
                    return
                result = func(*args)
                if result_func is not None:
                    GLib.idle_add(result_func, *result)
                GLib.idle_add(self.queue_size_notify_func, self.queue.qsize())
            except Exception:
                # keep worker running and report exceptions in the main thread
                GLib.idle_add(sys.excepthook, *sys.exc_info())
                
    def check_worker(self, in_worker):
        ct = current_thread()
        if (ct.name != 'MainThread') != in_worker:
            warnings.warn('check_worker({}) failed'.format(in_worker), stacklevel=2)
        
    def request(self, priority, func, args, result_func):
        self.queue.put_nowait((priority, (func, args, result_func)))
        self.queue_size_notify_func(self.queue.qsize())
        
    def clear(self, result_func):
        def clear_thread(queue):
            while True:
                try:
                    queue.get_nowait()
                except Queue.Empty:
                    return ()
        self.request(0, clear_thread, (self.queue,), result_func)
        
    def join(self):
        '''Request and wait for termination'''
        self.request(0, None, (), None)
        self.worker.join()
        
        
class LPTreeModel (GObject.Object):
    '''Treemodel for Launchpad objects

    Columns:    0:attrname,         type: str
                1:attrtype,         type: str (single char)
                2:value,            type: str
                3:value_visible,    type: bool
                4:row_visible,      type: bool
                5:icon_visible,     type: bool
                6:link_path,        type: object (tree path)
                7:is_linked,        type: bool
                8:attrname_visible, type: bool
                9:value_long,       type: str
               10:value_long_visible,type: bool
               11:stock_item,       type: str
               12:row_status,       type: str
               13:placeholder,      type: bool
    Treeiter:   tree path, type: (int,)*n
    '''
    column_func = [
        lambda self, tpath, row: '<i>{0}</i>'.format(row.attrname),#0
        lambda self, tpath, row: row.attrtype,#1
        lambda self, tpath, row: ' '.join(('' if row.attrtype not in 'AE' else
                        self.link_rows.get(row.self_link, dummy_row)[1].display_value
                            if row.state == ROW_LINK else
                            row.display_value).splitlines()),#2
        lambda self, tpath, row: (row.state != ROW_LOADING and row.attrtype in 'AE' and
                    not (row.long_value and Row.long_value)),#3
        lambda self, tpath, row: row.is_visible(),#4
        lambda self, tpath, row: row.state != ROW_OK,#5
        lambda self, tpath, row: (self.link_rows.get(row.self_link, dummy_row)[0]
                            if row.state == ROW_LINK
                            else tpath),#6
        lambda self, tpath, row: bool(row.self_link),#7
        lambda self, tpath, row: not row.long_value or not Row.long_value,#8
        lambda self, tpath, row: '<i>{0}:</i>\n{1}'
                            .format(row.attrname, GLib.markup_escape_text(row.display_value))
                            .strip(),#9
        lambda self, tpath, row: row.long_value and Row.long_value,#10
        lambda self, tpath, row: Row.row_icons[row.state],#11
        lambda self, tpath, row: ('Loading' if row.state == ROW_LOADING else
                            'Activate row to show attributes' if row.state == ROW_LINK else
                            (row.error.strip().splitlines() or [''])[0] if row.state == ROW_ERROR else
                            ''),#12
        lambda self, tpath, row: False,#13
    ]
    placeholder_data = ['<i>&lt;placeholder&gt;</i>','','',False,True,False,None,False,True,
                        '',False,'gtk-refresh','',True]
    column_type = [str,str,str,bool,bool,bool,object,bool,bool,str,bool,str,str,bool]
    assert len(column_func) == len(column_type) == len(placeholder_data)
    (COL_ATTRNAME, COL_ATTRTYPE, COL_VALUE, COL_VALUE_VISIBLE, COL_ROW_VISIBLE,
     COL_PIXBUF_VISIBLE, COL_LINK_PATH, COL_LINK_PATH_VISIBLE, COL_ATTRNAME_VISIBLE,
     COL_VALUE_LONG, COL_VALUE_LONG_VISIBLE, COL_STOCK_ITEM, COL_ROW_STATUS, COL_PLACEHOLDER,
    ) = range(len(column_func))
    
    __gsignals__ = {
        'row-changed-end': (GObject.SignalFlags.RUN_FIRST, None, (object, object)),
        'report-error': (GObject.SignalFlags.RUN_FIRST, None, (str,)),
    }
    
    cache_version = '2.18'
    
    
    ### (de)initialisation
    
    def __init__(self, lp_worker, cache_filename):
        GObject.Object.__init__(self)
        
        self.treestore = Gtk.TreeStore(*self.column_type)
        
        self.lp_worker = lp_worker
        self.cache_filename = cache_filename
        self.rows_info = {}  # key: treepath tuple, value: Row object
        self.root_row = None
        self.link_rows = {}  # key: self_link, value: (treepath tuple, Row object)
        
        data = self.load_cache()
        self.fill_data(data)
        try:
            self.test_rows()
        except AssertionError:
            sys.excepthook(*sys.exc_info())
            # discard corrupt cache
            self.fill_data(None)
            self.test_rows()
            
        self.create_root_children()

    def load_cache(self):
        try:
            with open(self.cache_filename, 'rb') as cache:
                cache_version = pickle.load(cache)
                data = pickle.load(cache) if cache_version == self.cache_version else None
        except Exception:
            sys.excepthook(*sys.exc_info())
            data = None
        return data
        
    def dump_cache(self):
        '''Write data to disk'''
        try:
            with open(self.cache_filename, 'wb') as cache:
                pickle.dump(self.cache_version, cache, -1)
                pickle.dump(self.get_data(), cache, -1)
        except (IOError, OSError, pickle.PickleError):
            sys.excepthook(*sys.exc_info())
            
    def create_root_children(self):
        if not self.root_row.n_children:
            self.create_entry_rows((), self.root_row)
            
        self.apply_rows((), None, with_children=True)
        
        #for row in self.gen_children():
        #    if row.state == ROW_LOADING:
        #        self.create_children(row)
                
    def fill_data(self, data):
        self.lp_worker.check_worker(False)
        if data:
            self.root_row = data[0]
            id_dict = data[1]
            for tpath, row in self.gen_BF():
                if row.parent is not None:
                    row.parent = id_dict[row.parent]
                if row.next is not None:
                    row.next = id_dict[row.next]
                if row.first_child is not None:
                    row.first_child = id_dict[row.first_child]
                self.rows_info[tpath] = row
                if row.self_link:
                    if row.state == ROW_OK:
                        self.link_rows[row.self_link] = tpath, row
                    elif row.state == ROW_LINK:
                        row.state = ROW_LOADING
            id_dict.clear()
        else:
            self.root_row = Row(self_link='root')
        self.link_rows[self.root_row.self_link] = (), self.root_row

        self.rows_info[()] = self.rows_info[None] = self.root_row
        self.root_row.value = self.lp_worker.launchpad
        self.root_row.attrname = 'lp'
        self.root_row.attrtype = 'E'

    def get_data(self):
        self.lp_worker.check_worker(False)
        return self.root_row, dict((id(row), row) for tpath, row in self.gen_BF())

    ### data iterators
    
    def gen_BF(self, tpath=(), row=None):
        '''iterate over all rows'''
        self.lp_worker.check_worker(False)
        if row is None:
            row = self.root_row
        rows = [(tpath, row)]
        while rows:
            tpath, row = rows.pop(0)
            while row:
                yield tpath, row
                rows.append((tpath+(0,), row.first_child))
                row = row.next
                if tpath:
                    tpath = tpath[:-1] +(tpath[-1]+1,)
                else:
                    assert row is None
                
    def gen_children(self, row=None):
        '''iterate over all children of a row'''
        self.lp_worker.check_worker(False)
        if row is None:
            row = self.root_row
        row = row.first_child
        while row:
            yield row
            row = row.next

    ### validation
    
    def test_row(self, row):
        self.lp_worker.check_worker(False)
        try:
            assert isinstance(row, Row)
            assert (row.parent is None) == (id(row.value) == id(self.lp_worker.launchpad))
            assert row.index >= 0
            assert row.next is None or row.index + 1 == row.next.index
            assert (row.parent is None or (row.next is None) == (
                    row.index + 1 == row.parent.n_children)), (row.parent, row)
            assert row.n_children >= 0
            assert (row.n_children == 0) == (row.first_child is None), (row,)
            assert row.first_child is None or row.first_child.parent == row
            assert row.first_child is None or row.first_child.index == 0, (row.first_child,)
            assert row.parent is None or row.parent.first_child is not None

            assert isinstance(row.attrname, str)
            assert row.attrtype in 'ACEO '
            if row.attrtype == 'C':
                assert row.value is None or isinstance(row.value, (
                                        str,
                                        lazr.restfulclient.resource.Collection)), type(row.value)
            elif row.attrtype == 'E':
                assert row.value is None or isinstance(row.value, (
                                        Launchpad,
                                        lazr.restfulclient.resource.Entry,
                                        lazr.restfulclient.resource.HostedFile,
                                        )), (row, type(row.value))
            elif row.attrtype == ' ':
                assert row.value is None, type(row.value)
            if row.state == ROW_ERROR:
                assert bool(row.error), (row,)
            if row.state == ROW_OK and row.self_link:
                assert row.self_link in self.link_rows, (row,)
            #FIXME:
            #if bool(row.self_link):
            #    link_row = self.link_rows[row.self_link][1]
            #    if row.state == ROW_LINK:
            #        assert link_row != row and link_row.state != ROW_LOADING, (row, link_row)
        except AssertionError as e:
            for arg in e.args:
                print(arg)
            raise

    def test_rows(self):
        self.lp_worker.check_worker(False)
        n = 0
        try:
            for tpath, row in self.gen_BF():
                self.test_row(row)
                n += 1
        finally:
            print('Tested %s rows' % n)

    ### public methods
    
    def new_filter(self, root):
        '''Return a filtermodel and attribute path for the root path'''
        filter_model = self.treestore.filter_new(root or None)
        filter_model.set_visible_column(self.COL_ROW_VISIBLE)
        attr_path = self.rows_info[root].get_attr_path()
        return filter_model, attr_path
        
    def refresh(self):
        '''Discard all data and populate with children of the root path'''
        self.lp_worker.check_worker(False)
        self.remove_children((), self.root_row)
        self.link_rows.clear()
        self.fill_data(None)
        self.create_root_children()
        
    def update_children(self, tpath):
        '''Make sure children exist and are up to date'''
        self.lp_worker.check_worker(False)
        row = self.rows_info[tpath]
        if row.state == ROW_LOADING:
            self.create_children(tpath, row)
        self.apply_rows(tpath, row, with_children=False)
        
    def reload_attributes(self):
        '''Update attribute-rows and rows marked loading or error'''
        self.lp_worker.check_worker(False)
        for tpath, row in self.gen_BF():
            if row.attrtype == 'A' or row.state in [ROW_LOADING, ROW_ERROR]:
                if row.state == ROW_ERROR:
                    row.value = None
                row.state = ROW_LOADING
                GLib.idle_add(self.on_replace_row, tpath, True)
        
    def reload_invisible_attrs(self):
        '''Update rows marked as invisible'''
        self.lp_worker.check_worker(False)
        for tpath, row in self.gen_BF():
            if row.attrname in Row.invisible_attrs and row.state == ROW_LOADING:
                GLib.idle_add(self.on_replace_row, tpath, True)
                
    def get_row_info(self, path):
        '''Returns info about the row for a path'''
        row = self.rows_info[path]
        return row.attrtype, row.attrname, row.value, row.state == ROW_LINK
        
    def get_attribute_path(self, path):
        '''Returns the attribute path for a row'''
        row = self.rows_info[path]
        return row.get_attr_path()
        
    def get_value_from_path(self, tpath):
        '''Returns the value of the value-column'''
        row = self.rows_info[tpath]
        return self.get_value_from_row(tpath, row, self.COL_VALUE)
        
    def get_status_text(self, tpath):
        '''Returns the value of the status-column'''
        row = self.rows_info[tpath]
        return self.get_value_from_row(tpath, row, self.COL_ROW_STATUS)
        
    def eval_operation(self, tpath, operation, kwargs):
        '''Execute Operation'''
        row = self.rows_info[tpath]
        self.eval_operation_main(tpath, row, operation, kwargs)
        
    ### internal methods
    
    def reload_entry_main(self, tpath, row):
        '''Update rows with a self_link but without value,
        also update collection rows without a self_link'''
        self.lp_worker.check_worker(False)
        if row.value is None and row.state != ROW_ERROR:
            assert row.self_link != 'root'
            if row.self_link:
                self.lp_worker.request(2, self.reload_entry_thread, (tpath, row),
                                          self.reload_entry_result)
            elif row.attrtype == 'C':
                self.get_row_value_main(2, tpath, row)
            else:
                row.error = 'No self_link available'
                row.state = ROW_ERROR
                self.on_replace_row(tpath, True)
                
    def reload_entry_thread(self, tpath, row):
        self.lp_worker.check_worker(True)
        try:
            row.value = self.lp_worker.launchpad.load(row.self_link)
        except HTTPError as e:
            row.error = e.content
            row.state = ROW_ERROR
        except Exception as e:
            row.error = str(e)
            row.state = ROW_ERROR
        else:
            self.finish_entry_row(tpath, row)
        return tpath,
        
    def reload_entry_result(self, tpath):
        self.lp_worker.check_worker(False)
        self.on_replace_row(tpath, True)
        
    def finish_entry_row(self, tpath, row):
        self.lp_worker.check_worker(True)
        try:
            row.self_link = row.value.self_link
        except Exception as e:
            row.error = str(e)
            row.state = ROW_ERROR
        else:
            if row.self_link not in self.link_rows:
                self.link_rows[row.self_link] = tpath, row
            row.state = ROW_LOADING

    def get_row_value_main(self, priority, tpath, row):
        self.lp_worker.check_worker(False)
        if row.parent.self_link or row.attrtype == 'O':
            if row.parent.value is None and row.parent.state != ROW_ERROR:
                assert row.parent.self_link != 'root'
                if row.parent.self_link:
                    self.lp_worker.request(priority, self.reload_entry_thread, (tpath[:-1], row.parent),
                                                     self.reload_entry_result)
                elif row.parent.attrtype == 'C':
                    self.get_row_value_main(priority, tpath[:-1], row.parent)
                    
        self.lp_worker.request(priority, self.get_row_value_thread, (tpath, row),
                                         self.get_row_value_result)
        
    def get_row_value_thread(self, tpath, row):
        '''update row value
        If value is None, row must be the child of an entry row.
        '''
        self.lp_worker.check_worker(True)
        try:
            if row.parent.attrtype == 'E' or row.attrtype == 'O':
                value = getattr(row.parent.value, row.attrname)
            else:
                value = None
        except HTTPError as e:
            row.error = e.content
            row.state = ROW_ERROR
        except Exception as e:
            row.error = str(e)
            row.state = ROW_ERROR
        else:
            row.value = value
            str_value = ''
            if row.attrtype == 'A':
                try:
                    str_value = str(value)
                except Exception as e:
                    row.error = str(e)
                    row.state = ROW_ERROR
                else:
                    row.state = ROW_OK
                    if row.attrname in Row.default_attrs and not row.parent.display_value:
                        row.parent.display_value = str_value
                        GLib.idle_add(self.on_replace_row, tpath[:-1], False)
            elif value is None:
                row.error = 'Value is None'
                row.state = ROW_ERROR
            elif row.attrtype == 'C':
                pass
            elif row.attrtype == 'E':
                self.finish_entry_row(tpath, row)
            elif row.attrtype == 'O':
                row.state = ROW_OK

            if row.state == ROW_LINK:
                try:
                    str_value = self.link_rows[row.self_link][1].display_value
                except KeyError:
                    pass
            row.long_value = row.value is not None and row.attrtype == 'A' and len(str_value) > 50
            row.display_value = str_value
            
        return tpath,
        
    def get_row_value_result(self, tpath):
        self.lp_worker.check_worker(False)
        self.on_replace_row(tpath, True)
        
    def get_collection_values_main(self, tpath, parent):
        '''Update children of a collection'''
        self.lp_worker.check_worker(False)
        row = parent.first_child
        idx = 0
        while row and row.attrtype != 'E':
            row = row.next
            idx += 1
        if row is None or row.value is not None or parent.value is None:
            return
        tpath += (idx,)
        self.lp_worker.request(row.priority(), self.get_collection_values_thread, (parent, tpath, row),
                                            self.get_collection_values_result)
        
    def get_collection_values_thread(self, parent, tpath, row):
        self.lp_worker.check_worker(True)
        try:
            coll_iter = iter(parent.value)
            entry_offset = row.index - 1
            while row:
                try:
                    value = coll_iter.next()
                except StopIteration:
                    row.attrname = '(empty)'
                    row.value = None
                    row.state = ROW_OK
                else:
                    row.attrname = str(row.index - entry_offset)
                    row.value = value
                    self.finish_entry_row(tpath, row)
                GLib.idle_add(self.on_replace_row, tpath, False)
                row = row.next
                tpath = tpath[:-1] + (tpath[-1]+1,)
            parent.state = ROW_OK
        except HTTPError as e:
            parent.error = e.content
            parent.state = ROW_ERROR
        return tpath[:-1],
        
    def get_collection_values_result(self, tpath):
        self.lp_worker.check_worker(False)
        self.on_replace_row(tpath, False)
        
    def remove_children(self, tpath, parent):
        '''removes all children of a row'''
        self.lp_worker.check_worker(False)
        gen = self.gen_BF(tpath, parent)
        gen.next()  # skip parent
        rows = []
        for tpath, row in gen:
            rows.append((row, tpath))
        for row, tpath in reversed(rows):
            row.delete()
            if row.self_link and self.link_rows.get(row.self_link, ((),None))[1] == row:
                del self.link_rows[row.self_link]
            self.on_remove_row(tpath)
        parent.n_children = 0
        parent.first_child = None

    def eval_operation_main(self, parent_tpath, parent, operation, kwargs):
        self.lp_worker.check_worker(False)
        row = parent.first_child
        tpath = parent_tpath + (0,)
        self.test_row(parent)
        self.test_row(row)
        
        parent.state = ROW_LOADING
        self.remove_children(tpath, row)
        row.attrname = '(empty)'
        row.attrtype = ' '
        row.value = None
        row.self_link = ''
        row.long_value = False
        row.display_value = ''
        row.state = ROW_LOADING
        row.error = ''
        self.test_row(parent)
        self.test_row(row)

        self.on_replace_row(tpath, True)
        self.on_replace_row(parent_tpath, True)
        
        self.lp_worker.request(1, self.eval_operation_thread,
                                     (parent_tpath, parent, tpath, row, operation, kwargs),
                                     self.eval_operation_result)
        
    def eval_operation_thread(self, parent_tpath, parent, tpath, row, operation, kwargs):
        self.lp_worker.check_worker(True)
        try:
            result = operation(**kwargs)
        except HTTPError as e:
            return str(e).splitlines()[0], None, None
        else:
            row.attrname = 'Result'
            if isinstance(result, lazr.restfulclient.resource.Entry):
                row.attrtype = 'E'
            elif isinstance(result, lazr.restfulclient.resource.HostedFile):
                row.attrtype = 'E'
            elif isinstance(result, lazr.restfulclient.resource.Collection):
                row.attrtype = 'C'
            else:
                row.attrtype = 'A'
            self.test_row(parent)
            self.test_row(row)
            row.value = result
            self.create_children(parent_tpath, parent)
            self.test_row(parent)
            self.test_row(row)
            return None, parent_tpath, tpath
        
    def eval_operation_result(self, errormsg, parent_path, row_path):
        self.lp_worker.check_worker(False)
        if errormsg is None:
            self.on_replace_row(row_path, True)
            self.on_replace_row(parent_path, True)
        else:
            self.emit('report-error', errormsg)
        
    def on_replace_row(self, tpath, has_child_toggled, with_children=False):
        '''Inform the model of the changes in the thread'''
        self.lp_worker.check_worker(False)
        row = self.rows_info[tpath]
        self.apply_rows(tpath, row, with_children=with_children)
        return False

    def on_remove_row(self, tpath):
        self.lp_worker.check_worker(False)
        if not tpath:
            return
        titer = self.treestore.get_iter(tpath)
        self.treestore.remove(titer)
        del self.rows_info[tpath]
        
    def create_row(self, prev_row, parent_tpath, parent, index, attrname, attrtype):
        '''Create a row, insert it in the hierarchy and update'''
        self.lp_worker.check_worker(False)
        row = Row(parent=parent, index=index,
                    attrname=attrname, attrtype=attrtype,
                    state=ROW_LOADING)
        tpath = (parent_tpath or ()) + (index,)
        if prev_row:
            prev_row.next = row
        else:
            parent.first_child = row
        self.rows_info[tpath] = row
        if row.is_visible() and attrtype == 'A' and attrname in Row.default_attrs:
            self.get_row_value_main(1, tpath, row)
        return row

    def create_entry_rows(self, parent_tpath, parent):
        self.lp_worker.check_worker(False)
        if parent.value is None:
            return
        try:
            link_tpath, link_row = self.link_rows[parent.self_link]
        except KeyError:
            return
        if parent != link_row:
            self.create_children(link_tpath, link_row)
            parent.state = ROW_LINK
            return
        lp_attributes = sorted(parent.value.lp_attributes)
        lp_collections = sorted(parent.value.lp_collections)
        lp_entries = sorted(parent.value.lp_entries)
        lp_operations = sorted(parent.value.lp_operations)
        count = sum(map(len, (lp_attributes, lp_collections, lp_entries, lp_operations)))
        i = 0
        parent.n_children = count
        row = None
        for attrtype, attrs in (
                        ('A', lp_attributes),
                        ('E', lp_entries),
                        ('C', lp_collections),
                        ('O', lp_operations)):
            for attrname in attrs:
                row = self.create_row(row, parent_tpath, parent, i, attrname, attrtype)
                self.on_replace_row(parent_tpath + (i,), False)
                i += 1
        parent.state = ROW_OK

    def create_collection_rows(self, parent_tpath, parent):
        self.lp_worker.check_worker(False)
        # list operations
        lp_operations = sorted(parent.value.lp_operations)
        count_O = len(lp_operations)
        row = None
        for i, attrname in enumerate(lp_operations):
            row = self.create_row(row, parent_tpath, parent, i, attrname, 'O')

        # list the first 50 entries
        count = count_O + 50
        parent.n_children = count
        for i in xrange(count_O, count):
            row = self.create_row(row, parent_tpath, parent, index=i, attrname='(empty)', attrtype='E')
        self.get_collection_values_main(parent_tpath, parent)

    def create_operation_row(self, parent_tpath, parent):
        self.lp_worker.check_worker(False)
        row = self.create_row(None, parent_tpath, parent, 0, '(empty)', ' ')
        parent.n_children = 1

    def create_children(self, tpath, row):
        '''Update row and create children if necessary'''
        self.lp_worker.check_worker(False)
        if row.state != ROW_LOADING:
            return
        if row.attrtype == 'A':
            self.get_row_value_main(row.priority(), tpath, row)
        elif row.attrtype == 'E':
            if row.n_children == 0:
                if row.value is None:
                    if row.self_link:
                        self.reload_entry_main(tpath, row)
                    else:
                        self.get_row_value_main(2, tpath, row)
                else:
                    self.create_entry_rows(tpath, row)
            else:
                row.state = ROW_OK
                self.reload_entry_main(tpath, row)
        elif row.attrtype == 'C':
            if row.value is None:
                self.get_row_value_main(row.priority(), tpath, row)
            else:
                if row.n_children == 0:
                    self.create_collection_rows(tpath, row)
                else:
                    self.get_collection_values_main(tpath, row)
        elif row.attrtype == 'O':
            if row.n_children == 0:
                self.create_operation_row(tpath, row)
            self.get_row_value_main(row.priority(), tpath, row)

    def apply_rows(self, tpath, row, with_children=False):
        self.lp_worker.check_worker(False)
        if with_children:
            rows = self.gen_BF(tpath, row)
        else:
            rows = [(tpath, row)]
        for tpath, row in rows:
            if not tpath:
                continue
            try:
                titer = self.treestore.get_iter(tpath)
            except ValueError:
                parent_path = tpath[:-1]
                parent_iter = self.treestore.get_iter(parent_path) if parent_path else None
                titer = self.treestore.insert(parent_iter, tpath[-1])
                if row.attrtype in 'ECO':
                    self.treestore.append(titer, self.placeholder_data)
            for i in range(len(self.column_func)):
                self.treestore.set_value(titer, i, self.get_value_from_row(tpath, row, i))
            self.emit('row-changed-end', tpath, titer)
                
    def get_value_from_row(self, tpath, row, column):
        self.lp_worker.check_worker(False)
        return self.column_func[column](self, tpath, row)
        

