#!/usr/bin/python
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

#  lp_browser: gtk-application using launchpadlib
#  Copyright © 2011, 2013  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function, division

import os, sys

from gi.repository import GLib
from gi.repository import Gtk
from gi.repository import Gdk
from gi.repository import Pango

from lp_browser.about import AboutDialog
from lp_browser.helpers import get_builder
from lp_browser.lpmodel import LPWorker, LPTreeModel, Row

from lp_browser import lplogin, wadl, selectservice
from lp_browser.operation import OperationDialog


class MainWindowHelper (object):
    def __init__(self):
        builder = get_builder('main.ui')
        self.window = builder.get_object('main_window')
        
        self.window.connect('destroy', self.on_destroy)
        builder.get_object('menuitem_about').connect('activate', self.on_about)
        builder.get_object('menuitem_quit').connect('activate', self.on_quit)

        toolbutton_refresh = builder.get_object('toolbutton_refresh')
        self.toolbutton_first = builder.get_object('toolbutton_first')
        self.toolbutton_back = builder.get_object('toolbutton_back')
        self.toolbutton_forw = builder.get_object('toolbutton_forw')
        self.toolbutton_last = builder.get_object('toolbutton_last')
        self.toolbutton_jumpto = builder.get_object('toolbutton_jumpto')
        self.toolbutton_execute = builder.get_object('toolbutton_execute')
        self.label_filter_path = builder.get_object('label_filter_path')
        self.treeview_main = builder.get_object('treeview_main')
        self.statusbar = builder.get_object('statusbar')
        self.label_status_service = builder.get_object('label_status_service')
        self.label_status_queue = builder.get_object('label_status_queue')
        
        menuitem_select_service = builder.get_object('menuitem_select_service')
        menuitem_select_service.connect('activate', self.on_menuitem_select_service_activate)
        
        column1 = builder.get_object('treeviewcolumn1')
        
        cell = Gtk.CellRendererText()
        cell.props.ellipsize_set = True
        cell.props.ellipsize = Pango.EllipsizeMode.END
        cell.props.single_paragraph_mode = True
        column1.pack_start(cell, True)
        column1.set_attributes(cell, text=2, visible=3)
        
        cell = Gtk.CellRendererText()
        cell.props.wrap_mode = Pango.WrapMode.WORD
        cell.props.wrap_width = 600
        column1.pack_start(cell, True)
        column1.set_attributes(cell, markup=9, visible=10)
        
        self.cell_value_long = cell
        self.column1 = column1
        
        toolbutton_refresh.connect('clicked', self.on_refresh)
        self.toolbutton_first.connect('clicked', self.on_toolbutton_first_clicked)
        self.toolbutton_back.connect('clicked', self.on_toolbutton_back_clicked)
        self.toolbutton_forw.connect('clicked', self.on_toolbutton_forw_clicked)
        self.toolbutton_last.connect('clicked', self.on_toolbutton_last_clicked)
        self.toolbutton_jumpto.connect('clicked', self.on_toolbutton_jumpto_clicked)
        self.toolbutton_execute.connect('clicked', self.on_toolbutton_execute_clicked)
        self.treeview_main.connect('cursor-changed', self.on_treeview_main_cursor_changed)
        self.treeview_main.connect('row-activated', self.on_treeview_main_row_activated)
        
        imagemenuitem_refresh = builder.get_object('imagemenuitem_refresh')
        imagemenuitem_refresh.connect('activate', self.on_refresh)
        imagemenuitem_refresh_all = builder.get_object('imagemenuitem_refresh_all')
        imagemenuitem_refresh_all.connect('activate', self.on_refresh_all)
        checkmenuitem_long_values = builder.get_object('checkmenuitem_long_values')
        checkmenuitem_long_values.connect('toggled', self.on_checkmenuitem_long_values_toggled)
        checkmenuitem_show_hidden = builder.get_object('checkmenuitem_show_hidden')
        checkmenuitem_show_hidden.connect('toggled', self.on_checkmenuitem_show_hidden_toggled)
        imagemenuitem_copy = builder.get_object('imagemenuitem_copy')
        imagemenuitem_copy.connect('activate', self.on_imagemenuitem_copy_activate)
        imagemenuitem_copyattr = builder.get_object('imagemenuitem_copyattr')
        imagemenuitem_copyattr.connect('activate', self.on_imagemenuitem_copyattr_activate)
        
        self.back_stack = []
        self.forw_stack = []
        self.current_root = ()
        self.method_info = {}
        
        self.treeview_main.grab_focus()
        
        self.model = None
        self.lp_worker = None
        self.treeview_main.connect('test-expand-row', self.on_test_expand_row)
        GLib.idle_add(self.on_idle_connect)
        GLib.idle_add(self.on_login, 'production')
        #GLib.idle_add(self.on_login, 'https://api.qastaging.launchpad.net/')
        
    def on_about(self, unused_widget):
        """Display the about box for lp-browser."""
        about = AboutDialog(self.window)
        response = about.dialog.run()
        about.dialog.destroy()

    def on_quit(self, unused_widget):
        """Signal handler for closing the main window."""
        self.window.destroy()

    def on_destroy(self, unused_widget):
        """Called when the main window is closed."""
        # Clean up code for saving application state should be added here.
        try:
            self.window.hide()
            while Gtk.events_pending():
                Gtk.main_iteration_do(False)
            print('stop worker and write cache ...', end=' ')
            sys.stdout.flush()
            self.clean_up()
            print('done.')
        finally:
            Gtk.main_quit()
            
    def on_idle_connect(self):
        self.treeview_main.connect_after("size-allocate", self.on_column1_size_allocate)
        
    def on_login(self, service):
        self.clean_up()
        self.method_info = {}
        launchpad, cache_filename, error = lplogin.login(service, self.window)
        method_cache_filename = os.path.splitext(cache_filename)[0] + '.methods'
        self.set_queue_status(0)
        self.set_status_text(error.splitlines()[0])
        self.lp_worker = LPWorker(launchpad, self.set_queue_status)
        if launchpad:
            self.method_info = wadl.get_method_info(launchpad._wadl.markup_url,
                                                       method_cache_filename)
            self.model = LPTreeModel(self.lp_worker, cache_filename)
            self.model.connect('row-changed-end', self.on_model_row_changed)
            self.model.connect('report-error', self.on_model_report_error)
        else:
            self.model = None
        self.back_stack = []
        self.forw_stack = []
        self.current_root = ()
        self.set_filter()
        self.label_status_service.set_text('Service: %s' % service)
        return False
        
    def clean_up(self):
        if self.lp_worker:
            self.lp_worker.join()
        if self.model:
            self.model.dump_cache()
            
    def on_column1_size_allocate(self, treeview, allocation):
        #FIXME: Because it is a TreeView, the width of a cell may vary for different lines.
        #col_pos = treeview.get_cell_area((0,), self.column1)
        cells = self.column1.props.cell_area.get_cells()
        #new_cell_width = self.column1.cell_get_position(self.cell_value_long)[1]
        #new_cell_width = allocation.width - self.column1.cell_get_position(cells[2])[0]
        new_cell_width = self.column1.cell_get_position(cells[2])[1] + self.column1.cell_get_position(cells[3])[1]
        if new_cell_width:
            self.cell_value_long.props.wrap_width = new_cell_width
            
    def get_cursor(self):
        current_path, col = self.treeview_main.get_cursor()
        if current_path:
            cell_area = self.treeview_main.get_background_area(current_path, col)
            visible_rect = self.treeview_main.get_visible_rect()
            x = visible_rect.x
            y = cell_area.y
            if y + cell_area.height > visible_rect.height:
                y = visible_rect.height - cell_area.height
            if y < 0:
                y = 0
        else:
            x, y = 0, 0
        return current_path, x, y
        
    def set_filter(self, current_path=None, x=0, y=0):
        if self.model:
            filter_model, attr_path = self.model.new_filter(self.current_root)
        else:
            filter_model = self.model
            attr_path = ''
        self.label_filter_path.set_text(attr_path)
        self.treeview_main.set_model(filter_model)
        if current_path:
            parent = current_path[:-1]
            if parent:
                self.treeview_main.expand_to_path(parent)
            col = self.treeview_main.get_column(0)
            self.treeview_main.set_cursor(current_path)
            cell_area = self.treeview_main.get_background_area(current_path, col)
            self.treeview_main.scroll_to_point(x, cell_area.y - y)
        self.toolbutton_first.set_sensitive(bool(self.back_stack))
        self.toolbutton_back.set_sensitive(bool(self.back_stack))
        self.toolbutton_forw.set_sensitive(bool(self.forw_stack))
        self.toolbutton_last.set_sensitive(bool(self.forw_stack))
        
    def on_refresh(self, unused_widget):
        if self.model:
            self.model.reload_attributes()
            
    def on_refresh_all(self, unused_widget):
        self.lp_worker.clear(self.on_refresh_all_finish)
        
    def on_refresh_all_finish(self):
        if self.model:
            self.model.refresh()
        if not self.lp_worker.launchpad:
            self.model = None
        self.back_stack = []
        self.forw_stack = []
        self.current_root = ()
        self.set_filter()
        
    def on_toolbutton_first_clicked(self, unused_widget):
        if not self.back_stack:
            return
        current_path, x, y = self.get_cursor()
        while self.back_stack:
            self.forw_stack.append((self.current_root, current_path, x, y))
            self.current_root, current_path, x, y = self.back_stack.pop()
        self.set_filter(current_path, x, y)
        
    def on_toolbutton_back_clicked(self, unused_widget):
        if not self.back_stack:
            return
        current_path, x, y = self.get_cursor()
        self.forw_stack.append((self.current_root, current_path, x, y))
        self.current_root, current_path, x, y = self.back_stack.pop()
        self.set_filter(current_path, x, y)
        
    def on_toolbutton_forw_clicked(self, unused_widget):
        if not self.forw_stack:
            return
        current_path, x, y = self.get_cursor()
        self.back_stack.append((self.current_root, current_path, x, y))
        self.current_root, current_path, x, y = self.forw_stack.pop()
        self.set_filter(current_path, x, y)
        
    def on_toolbutton_last_clicked(self, unused_widget):
        if not self.forw_stack:
            return
        current_path, x, y = self.get_cursor()
        while self.forw_stack:
            self.back_stack.append((self.current_root, current_path, x, y))
            self.current_root, current_path, x, y = self.forw_stack.pop()
        self.set_filter(current_path, x, y)
        
    def on_treeview_main_cursor_changed(self, treeview):
        path = treeview.get_cursor()[0]
        if path is None:
            status = ''
        else:
            model = treeview.get_model()
            status = model.get_value(model.get_iter(path), LPTreeModel.COL_ROW_STATUS)
        self.set_status_text(status)
        
    def on_toolbutton_jumpto_clicked(self, unused_widget):
        path = self.treeview_main.get_cursor()[0]
        self.do_action(self.treeview_main, path, execute=False)
        
    def on_toolbutton_execute_clicked(self, unused_widget):
        path = self.treeview_main.get_cursor()[0]
        self.do_action(self.treeview_main, path, execute=True)
        
    def on_treeview_main_row_activated(self, treeview, path, unused_view_column):
        self.do_action(treeview, path, execute=None)
        
    def on_test_expand_row(self, treeview, titer, path):
        path = treeview.get_model().convert_path_to_child_path(path)
        path = tuple(path.get_indices())
        self.model.update_children(path)
        return False  # allow expansion
        
    def do_action(self, treeview, path, execute=False):
        '''Perform an action on a row
        
        execute: kind of action, possible values:
                 False: follow link if possible
                 True: operation dialog if possible
                 None: operation dialog for operation rows, follow link for other rows
        '''
        filtermodel = treeview.get_model()
        if not path:
            return
        titer = filtermodel.get_iter(path)
        link_path = filtermodel.get_value(titer, self.model.COL_LINK_PATH)
        path = filtermodel.convert_path_to_child_path(path)
        path = tuple(path.get_indices())
        
        attrtype, attrname, value, islink = self.model.get_row_info(path)
        if attrtype == 'O' and execute != False:
            if value is None:
                return
            dialog = OperationDialog(self.window)
            if dialog.set_operation(attrname, value, self.method_info):
                result = dialog.dialog.run()
                if result == Gtk.ResponseType.OK:
                    try:
                        kwargs = dialog.kwargs
                    except AttributeError as e:
                        print(e)
                    else:
                        self.model.eval_operation(path, dialog.operation, kwargs)
            dialog.dialog.destroy()
        elif execute == True:
            pass
        elif filtermodel.iter_has_child(titer) or islink:
            current_path, x, y = self.get_cursor()
            self.back_stack.append((self.current_root, current_path, x, y))
            del self.forw_stack[:]
            self.current_root = link_path or path
            self.set_filter()
        
    def on_checkmenuitem_long_values_toggled(self, widget):
        Row.long_value = widget.get_active()
        self.treeview_main.columns_autosize()
        
    def on_checkmenuitem_show_hidden_toggled(self, widget):
        Row.show_hidden = widget.get_active()
        self.set_filter()
        self.model.reload_invisible_attrs()
        
    def on_imagemenuitem_copy_activate(self, unused_widget):
        if self.model is None:
            return
        clipboard = Gtk.Clipboard.get(Gdk.Atom.intern('CLIPBOARD', False))
        path = self.treeview_main.get_cursor()[0]
        if path is None:
            path = self.current_root
        else:
            filtermodel = self.treeview_main.get_model()
            path = filtermodel.convert_path_to_child_path(path)
            path = tuple(path.get_indices())
        textdata = self.model.get_value_from_path(path)
        clipboard.set_text(textdata, len(textdata)+1)
        
    def on_imagemenuitem_copyattr_activate(self, unused_widget):
        if self.model is None:
            return
        clipboard = Gtk.Clipboard.get(Gdk.Atom.intern('CLIPBOARD', False))
        path = self.treeview_main.get_cursor()[0]
        if path is None:
            path = self.current_root
        else:
            filtermodel = self.treeview_main.get_model()
            path = filtermodel.convert_path_to_child_path(path)
            path = tuple(path.get_indices())
        textdata = self.model.get_attribute_path(path)
        clipboard.set_text(textdata, len=-1)
        
    def set_status_text(self, message):
        self.statusbar.pop(self.statusbar.get_context_id('X'))
        self.statusbar.push(self.statusbar.get_context_id('X'), message)
        
    def on_model_row_changed(self, model, path, titer):
        cursor_path = self.treeview_main.get_cursor()[0]
        if cursor_path is None:
            return
        cursor_path = tuple(cursor_path.get_indices())
        if path == cursor_path:
            status = self.model.get_status_text(path)
            self.set_status_text(status)
            
    def set_queue_status(self, count):
        status_text = '{0} pending items'.format(count)
        # This idle thing prevents errors/warnings in Gtk/GLib/GObject/Pango
        GLib.idle_add(self.on_idle_set_queue_status, status_text)
        
    def on_model_report_error(self, unused_widget, message):
        self.set_status_text(message)
        
    def on_idle_set_queue_status(self, status_text):
        self.label_status_queue.set_text(status_text)
        
    def on_menuitem_select_service_activate(self, unused_widget):
        service = selectservice.select_service(self.window)
        if service is not None:
            GLib.idle_add(self.on_login, service)
            
def main():
    main_window = MainWindowHelper()
    main_window.window.show()
    GLib.threads_init()
    Gtk.main()
    
