#!/usr/bin/python
# -*- coding:utf-8 -*-

#  Copyright © 2011, 2013  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function, division

import sys
import httplib2
from xml.dom.minidom import parseString
import cPickle as pickle


method_info = {}
application_infos = []


def get_wadl_file(url):
    http = httplib2.Http()
    headers = {'Accept': 'application/vd.sun.wadl+xml'}
    response, content = http.request(url, headers=headers)
    return content if response.status == 200 else None

def parse_wdi(wo, indent=0):
    sindent = '  ' * indent
    print('%snodeName=%s, nodeType=%s, nodes=%s' % (sindent, wo.nodeName, wo.nodeType, len(wo.childNodes)), end=' ')
    print(wo.data.strip()[:20] if wo.nodeName == '#text' else '')
    if wo.attributes:
        for k, v in wo.attributes.items():
            if k not in ['xmlns:xsi', 'xmlns', 'xmlns:xsd', 'xmlns:wadl', 'xsi:schemaLocation']:
                print('  %sa %s=%s' % (sindent, k, v))

    for i, n in enumerate(wo.childNodes):
        if i == 30:
            break
        parse_wdi(n, indent+1)

def parse_document(wo):
    for n in wo.childNodes:
        if n.nodeName == 'wadl:application':
            parse_application(n)
            break

def parse_application(wo):
    for n in wo.childNodes:
        if n.nodeName == 'wadl:doc':
            title, descr_short, descr_long, descr_extra = parse_doc(n)
            application_infos.append((title, '\n'.join((descr_short, descr_long, descr_extra))))
        elif n.nodeName == 'wadl:resource_type':
            parse_resource_type(n)

def parse_resource_type(wo):
    resource_id = wo.attributes.get('id')
    resource_id = resource_id and resource_id.value
    for n in wo.childNodes:
        #if n.nodeName == 'wadl:doc':
        #    title, descr_short, descr_long, descr_extra = parse_doc(n)
        if n.nodeName == 'wadl:method':
            parse_method(n)

def parse_method(wo):
    method_id = wo.attributes['id'].value
    method_name = wo.attributes['name'].value
    descr_short, descr_long = ('', '')
    params = []
    for n in wo.childNodes:
        if n.nodeName == 'wadl:doc':
            assert (descr_short, descr_long) == ('', '')
            unused_title, descr_short, descr_long, unused_descr_extra = parse_doc(n)
        elif n.nodeName == 'wadl:request':
            assert params == []
            params = parse_request(n)
    method_info[method_id] = (method_name, descr_short, descr_long, params)

def parse_request(wo):
    params = []
    for n in wo.childNodes:
        if n.nodeName == 'wadl:param':
            params.append(parse_param(n))
        elif n.nodeName == 'wadl:representation':
            assert params == []
            params = parse_representation(n)
    return params

def parse_representation(wo):
    params = []
    for n in wo.childNodes:
        if n.nodeName == 'wadl:param':
            params.append(parse_param(n))
    return params

def parse_param(wo):
    #style = wo.attributes['style'].value
    required = wo.attributes.get('required')
    required = (required and required.value.lower()) == u'true'
    name = wo.attributes['name'].value
    descr_short = descr_long = ''
    for n in wo.childNodes:
        if n.nodeName == 'wadl:doc':
            assert not (descr_short or descr_long), (descr_short, descr_long)
            unused_title, descr_short, descr_long, unused_descr_extra = parse_doc(n)
            assert descr_short or descr_long, (descr_short, descr_long)
    return name, required, descr_short, descr_long

def parse_doc(wo):
    title = wo.attributes.get('title')
    title = title and title.value
    text = []
    texts = []
    for n in wo.childNodes:
        if n.nodeName == '#text':
            text += n.data.splitlines()
        if n.nodeName == 'p':
            texts.append(parse_p(n))
    text = ' '.join(t for t in (t.strip() for t in text) if t)
    descr_short = texts.pop(0) if texts else ''
    descr_long = texts.pop(0) if texts else ''
    descr_extra = '\n'.join(texts)
    assert not text == bool(descr_short or descr_long), (text, descr_short, descr_long)
    if text:
        if len(text) > 35:
            return title, '', text, ''
        else:
            return title, text, '', ''
    else:
        return title, descr_short, descr_long, descr_extra

def parse_p(wo):
    texts = []
    for n in wo.childNodes:
        if n.nodeName == '#text':
            texts += n.data.splitlines()
        elif n.nodeName == 'a':
            texts += parse_a(n)
    text = ' '.join(t for t in (t.strip() for t in texts) if t)
    return text

def parse_a(wo):
    link = wo.attributes['href']
    #text = []
    #for n in wo.childNodes:
    #    if n.nodeName == '#text':
    #        text += n.data.splitlines()
    return [link.value]

def parse_wadl_file(content):
    doc = parseString(content)
    method_info.clear()
    #parse_wdi(doc)
    parse_document(doc)

cache_version = '1.2'

def get_method_info(url, cache_filename):
    try:
        with open(cache_filename, 'rb') as cache:
            cache_version_, url_, data = pickle.load(cache)
            if cache_version_ != cache_version or url_ != url:
                data = None
    except (IOError, OSError, ValueError, EOFError, pickle.PickleError):
        data = None
    method_info.clear()
    if data is not None:
        method_info.update(data)
    else:
        wadl = get_wadl_file(url)
        if wadl is not None:
            parse_wadl_file(wadl)
            try:
                with open(cache_filename, 'wb') as cache:
                    pickle.dump((cache_version, url, method_info), cache, -1)
            except (IOError, OSError, pickle.PickleError):
                pass
    return method_info

def _main():
    args = sys.argv[1:]
    arg1 = args[0] if args else '0'
    url = 'https://api.qastaging.launchpad.net/1.0'
    cache = './' + url.replace('/', '-') + '.methods'
    wadlfile = './' + url.replace('/', '-') + '.wadl'
    if arg1 in '01':
        wadl = get_wadl_file(url)
        print()
        if wadl is None:
            print('No wadl')
        else:
            open(wadlfile, 'wt').write(wadl)
            print('wadlfile written to', wadlfile)
            print()
            if arg1 == '0':
                parse_wdi(wadl)
            else:
                parse_wadl_file(wadl)
    else:
        print('cache in', cache)
        print()
        mi = get_method_info(url, cache)
        assert mi == method_info
    #print
    #print '### method_info ###'
    #for mid, (method_name, descr_short, descr_long, params) in method_info.items():
    #    print ' ', mid, method_name, descr_short, descr_long

if __name__ == '__main__':
    _main()


